function [symbol reg_dec Dynamic_Table] = mq_decoder(label,reg_dec,Dynamic_Table,Static_Table)

%     This file is part of mq_codec.
% 
%     mq_codec is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     any later version.
% 
%     mq_codec is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with mq_codec.  If not, see <http://www.gnu.org/licenses/>.
%
% MQ decoder routine
% 
% Syntax
%     [symbol reg_enc Dynamic_Table] = mq_decoder(label,reg_enc,...
%         Dynamic_Table,Static_Table)
%     
% Inputs
%     label ~ label for that symbol
%     reg_dec ~ state-registers for the coder (before decode)
%     Dynamic_Table ~ arithmetic code states table (before decode)
%     Static_Table ~ transition table (see p. 75 of [1])
%     
% Outputs
%     symbol ~ decoded symbol
%     reg_dec ~ state-registers for the coder (after decode)
%     Dynamic_Table ~ arithmetic code states table (after decode)
%     
% Note
%     Implemented from p. 483 of [1]
% 
% [1] David S. Taubman and Michael W. Marcellin, JPEG2000, Image Compression 
%     Fundamentals, Standards and Practice, Kluwer Academic Publishers, 2002
%
% see also mq_decoder_init, mq_renormalize_once, mq_encoder
%
% Xavier Alameda Pineda (xavi.alameda at gmail.com) v1.0 - 20/May/2009
% Image and Video Processing Group, Technical University of Catalonia

if nargin < 4
    error('4 arguments needed.');
end

% C subsets indexes

C_active  =  9:24;
C_active_mask  =  uint32(sum(2.^(C_active-1)));


% % % ===============START
% % % MQ Decoder Procedure
% % % ===============START

% Get information related to the label
expected_symbol = Dynamic_Table(label,2);               % s = s_k
probability = Static_Table(Dynamic_Table(label,1),4);   % p = p(sigma_k)

if expected_symbol > 1
    disp('pstlasdkldf');
end

reg_dec.A = reg_dec.A - uint16(probability);

if reg_dec.A < uint16(probability),
    % Conditional exchange of MPS and LPS
    expected_symbol = 1 - expected_symbol;
end

% Compare active reg_decion of C
if bitshift(bitand(reg_dec.C,C_active_mask),-(min(C_active)-1)) < probability
    symbol = 1-expected_symbol;
    reg_dec.A = uint16(probability);
else
    symbol = expected_symbol;
    Temp = bitshift(bitand(reg_dec.C,C_active_mask),-(min(C_active)-1))-uint32(probability);
    reg_dec.C = bitand(reg_dec.C,bitcmp(C_active_mask));
    reg_dec.C = reg_dec.C + bitand(uint32(bitshift(Temp,min(C_active)-1)),C_active_mask);
end

if reg_dec.A < 2^15,
    % The symbol was a real MPS
    if symbol == Dynamic_Table(label,2),
        Dynamic_Table(label,1) = Static_Table(Dynamic_Table(label,1),1);
    % The symbol was a real LPS
    else
        Dynamic_Table(label,2) = bitxor(Dynamic_Table(label,2),...
                                        Static_Table(Dynamic_Table(label,1),3));
        Dynamic_Table(label,1) = Static_Table(Dynamic_Table(label,1),2);
    end
end

% Perform quantization shift
while reg_dec.A < 2^15,
    reg_dec = mq_renormalize_once(reg_dec);
end

% % % =================END
% % % MQ Decoder Procedure
% % % =================END

return