function [reg_enc Dynamic_Table] = mq_encoder(symbol,label,reg_enc,Dynamic_Table,Static_Table)

%     This file is part of mq_codec.
% 
%     mq_codec is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     any later version.
% 
%     mq_codec is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License
%     along with mq_codec.  If not, see <http://www.gnu.org/licenses/>.
%
% MQ encoder routine
% 
% Syntax
%     [reg_enc Dynamic_Table] = mq_encoder(symbol,label,reg_enc,...
%         Dynamic_Table,Static_Table)
%     
% Inputs
%     symbol ~ symbol to code
%     label ~ label for that symbol
%     reg_enc ~ state-registers for the coder (before code)
%     Dynamic_Table ~ arithmetic code states table (before code)
%     Static_Table ~ transition table (see p. 75 of [1])
%     
% Outputs
%     reg_enc ~ state-registers for the coder (after code)
%     Dynamic_Table ~ arithmetic code states table (after code)
%     
% Note
%     Implemented from p. 478 of [1]
% 
% [1] David S. Taubman and Michael W. Marcellin, JPEG2000, Image Compression 
%     Fundamentals, Standards and Practice, Kluwer Academic Publishers, 2002
%
% see also mq_encoder_init, mq_transfer_byte, mq_decoder, mq_codec_init_tables
%
% Xavier Alameda Pineda (xavi.alameda at gmail.com) v1.0 - 20/May/2009
% Image and Video Processing Group, Technical University of Catalonia

if nargin < 5
    error('5 arguments needed.');
end

% % % ===============START
% % % MQ Encoder Procedure
% % % ===============START

% Get information related to the label
expected_symbol = Dynamic_Table(label,2);               % s = s_k
probability = Static_Table(Dynamic_Table(label,1),4);   % p = p(sigma_k)

reg_enc.A = reg_enc.A - probability;

if reg_enc.A < probability,
    % Conditional exchange of MPS and LPS
    expected_symbol = 1 - expected_symbol;
end

if symbol == expected_symbol,
    % Assign MPS the upper sub-interval
    reg_enc.C = reg_enc.C + uint32(probability);
else
    % Assign LPS the lower sub-interval
    reg_enc.A = probability;
end

if reg_enc.A < 2^15,
    % The symbol was a real MPS
    if symbol == Dynamic_Table(label,2),
        Dynamic_Table(label,1) = Static_Table(Dynamic_Table(label,1),1);
    % The symbol was a real LPS
    else
        Dynamic_Table(label,2) = bitxor(Dynamic_Table(label,2),...
                                        Static_Table(Dynamic_Table(label,1),3));
        Dynamic_Table(label,1) = Static_Table(Dynamic_Table(label,1),2);
    end
end

% Perform quantization shift
while reg_enc.A < 2^15,
    reg_enc.A = 2*reg_enc.A;
    reg_enc.C = 2*reg_enc.C;
    reg_enc.t = reg_enc.t-1;
    if reg_enc.t == 0,
        reg_enc = mq_transfer_byte(reg_enc);
    end
end

% % % =================END
% % % MQ Encoder Procedure
% % % =================END